"use strict";
exports.__esModule = true;
var Future = /** @class */ (function () {
    function Future(action) {
        this.action = action;
    }
    /**
     * Start execution of the Future. Accepts resolve/reject parameters
     * @param {Function} reject  Handler if error occured during Future execution
     * @param {Function} resolve Handler if Future fully executed successfully
     */
    Future.prototype.engage = function (reject, resolve) {
        var potentialPromise;
        try {
            //In the case where the action call just completely blows up, prevent against that and invoke reject.
            potentialPromise = this.action(reject, resolve);
        }
        catch (error) {
            reject(error);
        }
        if (potentialPromise && typeof potentialPromise.then === 'function') {
            potentialPromise.then(resolve)["catch"](reject);
        }
    };
    /**
     * Similar to engage. Starts execution of the Future and returns the resolve/reject wrapped up in a Promise instead of taking reject/resolve parameters
     * @return {Promise<ResultType>} Start execution of the Future but return a Promise which will be resolved/reject when the Future is
     */
    Future.prototype.toPromise = function () {
        var _this = this;
        return new Promise(function (resolve, reject) { return _this.engage(reject, resolve); });
    };
    /**
     * Modify the data within the pipeline synchronously
     * @param {Function} mapper Method which will receive the current data and map it to a new value
     */
    Future.prototype.map = function (mapper) {
        return this.flatMap(function (x) { return Future.of(mapper(x)); });
    };
    /**
     * Run another asynchronous operation recieving the data from the previous operation
     * @param {Function} next Method to execute to run the operation
     */
    Future.prototype.flatMap = function (next) {
        var _this = this;
        return new Future(function (reject, resolve) {
            _this.engage(reject, function (data) { return next(data).engage(reject, resolve); });
        });
    };
    /**
     * Attempt to recover from an error in the pipeline in order to continue without rejecting the Future. The repaired type must extend
     * the original ResultType as to make sure follow-on Futures can handle the data further in the chain
     * @param {Function} errHandler Error handler which should return a new Future and resolve or reject result
     */
    Future.prototype.handleWith = function (errHandler) {
        var _this = this;
        return new Future(function (reject, resolve) {
            _this.engage(function (error) {
                errHandler(error).engage(reject, resolve);
            }, resolve);
        });
    };
    /**
     * Map errors to a new error type.
     * @param {Function} mapper Mapping function which will recieve the current error can map it to a new type
     */
    Future.prototype.errorMap = function (mapper) {
        var _this = this;
        return new Future(function (reject, resolve) {
            _this.engage(function (error) { return reject(mapper(error)); }, resolve);
        });
    };
    /**
     * Create a new synchronous Future which will automatically resolve with the provided value
     */
    Future.of = function (result) {
        return new Future(function (_, resolve) {
            resolve(result);
        });
    };
    /**
     * Create a new synchronous Future which will automatically reject with the provided value
     */
    Future.reject = function (error) {
        return new Future(function (reject) {
            reject(error);
        });
    };
    /**
     * Returns a new Future which will run the two provided futures in "parallel". The returned Future will be resolved if both
     * of the futures resolve and the results will be in an array properly indexed to how they were passed in. If any of the Futures
     * reject, then no results are returned and the Future is rejected.
     */
    Future.gather2 = function (future1, future2) {
        return new Future(function (reject, resolve) {
            var results = [];
            var count = 0;
            var done = false;
            future1.engage(function (error) {
                if (!done) {
                    done = true;
                    reject(error);
                }
            }, function (result) {
                results[0] = result;
                if (++count === 2) {
                    resolve(results);
                }
            });
            future2.engage(function (error) {
                if (!done) {
                    done = true;
                    reject(error);
                }
            }, function (result) {
                results[1] = result;
                if (++count === 2) {
                    resolve(results);
                }
            });
        });
    };
    /**
     * Same as gather2 except supports running three concurrent Futures
     */
    Future.gather3 = function (future1, future2, future3) {
        var firstTwo = this.gather2(future1, future2);
        return this.gather2(firstTwo, future3).map(function (_a) {
            var _b = _a[0], f1 = _b[0], f2 = _b[1], f3 = _a[1];
            return [f1, f2, f3];
        });
    };
    /**
     * Same as gather2 except supports running four concurrent Futures
     */
    Future.gather4 = function (future1, future2, future3, future4) {
        var firstTwo = this.gather2(future1, future2);
        var secondTwo = this.gather2(future3, future4);
        return this.gather2(firstTwo, secondTwo).map(function (_a) {
            var _b = _a[0], f1 = _b[0], f2 = _b[1], _c = _a[1], f3 = _c[0], f4 = _c[1];
            return [f1, f2, f3, f4];
        });
    };
    /**
     * Returns a new Future which will run all of the provided futures in parallel. The returned Future will be resolved if all
     * of the futures resolve and the results will be in an array properly indexed to how they were passed in. If any of the Futures
     * reject, then no results are returned
     */
    Future.all = function (futures) {
        return new Future(function (reject, resolve) {
            var results = [];
            var count = 0;
            var done = false;
            futures.forEach(function (futureInstance, index) {
                futureInstance.engage(function (error) {
                    if (!done) {
                        done = true;
                        reject(error);
                    }
                }, function (result) {
                    results[index] = result;
                    count += 1;
                    if (count === futures.length) {
                        resolve(results);
                    }
                });
            });
        });
    };
    return Future;
}());
exports["default"] = Future;
